/*
 * VerletWeis 
 * Copyright 2004 Giuseppe Foffi & Gavin McCullagh
 * www_gmc@fiachra.ucd.ie
 */
/*
 * This file is part of VerletWeis.
 * VerletWeis is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * VerletWeis is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VerletWeis; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

// See Physical Review A  Vol 5  Num 2  pp 939-952  
// Feb 1972

#include "verletweis.h"

/* getGvector added by Gavin */


VerletWeis::VerletWeis(double eta, double R_start, double R_finish, double R_delta, double sigma) 
{
  this->eta = eta;
  this->R_start = R_start;
  this->R_finish = R_finish;
  this->R_delta = R_delta;
  this->sigma = sigma;
}

// density at close packing (0.74048049)
// density in this case.

int VerletWeis::getVWGvector(vector<double> &r, vector<double> &gr)
{
    // assure vector is empty.
	r.clear();
	gr.clear();

    // set values for r < sigma
	for(double R = R_start; R < sigma; R += R_delta) { 
		r.push_back(R);
		gr.push_back(0);
	}

	for(double R = this->sigma; R <= this->R_finish; R += this->R_delta) { 
		r.push_back(R);
		gr.push_back(this->g_radial_verlet(R,eta));
	}

	return(0);
}

int VerletWeis::getPYVWGvector(vector<double> &r, vector<double> &vwgr, vector<double> &pygr)
{
    // assure vector is empty.
	r.clear();
	vwgr.clear();
	pygr.clear();

    // set values for r < sigma
	for(double R = R_start; R < sigma; R += R_delta) { 
		r.push_back(R);
		vwgr.push_back(0);
		pygr.push_back(0);
	}

	for(double R = this->sigma; R <= this->R_finish; R += this->R_delta) { 
		r.push_back(R);
		vwgr.push_back(this->g_radial_verlet(R,eta));
		pygr.push_back(this->g_radial(R,eta));
	}

	return(0);
}

/*

pretty clear get the distance r and the packing fraction phi and 
return the radial distribution there.
in the header there are prototypes it seems you need them.

*/

// see eqn 2.6
double VerletWeis::g_radial_verlet(const double& r, const double& phi) 
{
  double phip = phi - pow(phi,2)/16;            // see eqn 2.9
  double dprime = pow(phip/phi, (1./3.));       
  return g_radial(r/dprime, phip) + deltaG(r,phip) ;
}

// see eqn 2.7
double VerletWeis::deltaG(const double &r, const double &p) 
{
  double expo = exp( -alpha(p) * (r-1) );
  double coso = r*cos(alpha(p) * (r-1));
  return A(p) * expo * coso; 
}

// see eqn 2.12
double VerletWeis::A( const double& e) 
{
  double num = e * e * (1 - 0.7117 * e - 0.114 * pow( e, 2));
  double den = pow((1 - e ), 4 );
  return (3 * num) / (4 * den);
}

// see eqn 2.17 (alpha  ==  mu * d) (e = eta_w)
double VerletWeis::alpha( const double& e ) 
{
  double num = A(e) * 24;
  double den = e * g_radial(1, e);
  return num/den;
}

double VerletWeis::g_radial(double r, double p) 
{
  return  Heavi(r-1.)*G_1(r, p).real() + Heavi(r-2.)*G_2(r, p).real();
}

/******************************************************************************
 *  This Functions collection generates G_n to get PY radial distribution     *
 *  function.For more dettails see Gast and Russel (G&R) or Simple            *
 *  Liquids pag.125                                                           *
 *  and around there...                                                       *
 *  All the functions are inside this file and you    just need  the G_n      *
 *  functions                                                                 *
 *  PARAMETER: p ---> packing fraction                                        *
 *             r ---> distance                                                *
 *  OUTPUT: Functions to generates g(r) PY. see Gast eq 16 -> 19              *
 *  DATE:10/7/2001                                                            *
 *                                                                            *
 *****************************************************************************/

//Heaviside function to costruct g(r) eq 16 G&R
double VerletWeis::Heavi(const double &x) 
{
  if(x >= 0) 
    return 1.;
  else
    return 0.;
}

// Function G_1 as from 17 G&R
complex <double> VerletWeis::G_1(const double& r, const double& p) 
{
  vector < complex<double> > ti(3);
  Spoles(p, ti);
  complex <double> term(0,0);
  complex <double> deno;
  complex <double> nume;
  // This loop is for cycle permutation on indeces
  for(int i = 0; i < 3; ++i) {
    nume = (ti[i]*L(p,ti[i]))*exp(ti[i]*(r-1));
    deno = ((ti[i]-ti[((i+1)%3)])*(ti[i]-ti[((i+2)%3)]));
    term += nume/deno;
  }
  return term/r/ pow((1-p),2);
} 

// Function G_2 as from 17 G&R
complex <double> VerletWeis::G_2(const double& r, const double& p) 
{
  vector < complex <double> > ti(3);
  Spoles(p, ti);
  complex <double> one, two, three, tmp1, tmp3(0,0); //,tmp2;
  complex <double> term(0,0);
  for(int i = 0; i < 3; ++i) {
    
    tmp1 = L(p,ti[i])/((ti[i]-ti[((i+1)%3)])*(ti[i]-ti[((i+2)%3)]));
    one = pow(tmp1, 2) * exp(ti[i]*(r-2));
    two = pow(tmp1, 2) * exp(ti[i]*(r-2)) * ti[i]*(r-2);
    tmp3 = (Lprime(p,ti[i])*((ti[i]-ti[((i+1)%3)])*(ti[i]-ti[((i+2)%3)])));
    tmp3 -= L(p,ti[i])*((ti[i]-ti[((i+1)%3)])+(ti[i]-ti[((i+2)%3)]));
    tmp3 /= pow((ti[i]-ti[((i+1)%3)])*(ti[i]-ti[((i+2)%3)]),2);
    three = ti[i] * exp(ti[i]*(r-2)) *2.0 * tmp1*tmp3;
    term += one +two + three;
    
  }

  return -12*p*term/r/ pow((1-p),2)/ pow((1-p),2);
}

// complex <double> G_3(const double& r, const double& p) {
//   vector < complex <double> > ti(3);
//   Spoles(p, ti);
//   complex <double> factor, add;
//   for(int i = 0; i < 3; ++i) {
//     int u = ((i+1)%3);
//     int d = ((i+2)%3);
//     factor = exp(ti[i]*(r-3));
//     factor *= (L(p,ti[i]));
//     factor /= pow((ti[i]-ti[u])*(ti[i]-ti[d]),3);
//     add += 6*(L(p,ti[i]))*Lprime(p,ti[i]);
//     add -=  6*pow(L(p,ti[i]),2)/(ti[i]-ti[u]);
  
//   }
// }
/*                  STATIC FROM HERE             */


//This func. takes packing fraction p and returns ( by reference)
//the poles of the equation S(t) eq18 
void VerletWeis::Spoles(const double & p, vector< complex <double> > & poles) 
{
  vector<double> coef(4); 
  //inizialize the coefficients of S(t) as in eq. 18 in Gast
  double local_tmp = pow((1-p),2);
  coef[3] = 1;
  coef[2] = 6*p*(1-p)/local_tmp;
  coef[1] = 18*p*p/local_tmp;
  coef[0] = -12*p*(1+2*p)/local_tmp; 
  Solve3(coef, poles);
  
}

//Func L(t) as from eq 19
complex<double> VerletWeis::L(const double & p, const complex<double>& t) 
{
  double a1 = (1 + .5 * p);
  double a0 = (1 + 2 * p);
  return a1*t+a0;
}

//Derivative of L(t) [TRIVIAL]
double VerletWeis::Lprime(const double & p, const complex <double>& t) 
{
  double a1 = (1 + .5 * p);
  //double a0 = (1 + 2 * p);
  return a1;
}

//
//Given the vector  a[] of
//      a[3]*x^3+a[2] x^2 +a[1] x +a[0]
//
// returns by reference in sol the real part of the solution of the equation
// returns 1 if at leas one of the root is complex.
//
int VerletWeis::Solve3(const vector<double> & a, vector <double> & sol) 
{
  vector< complex<double> > sol_t(3);
  double tmp1 = -a[2]/(3*a[3]);
  double tmp2 = 1/(3*pow(2,(1./3.))*a[3]);
  double tmp3 = pow(2,(1./3.))*(-a[2]*a[2]+3*a[3]*a[1]);
  complex <double> number1(1, sqrt(3.0));
  complex <double> number2(1, -sqrt(3.0));
  sol_t[0] = tmp1-tmp3*(3*a[3]*getTerm(a))+tmp2*getTerm(a);
  
  sol_t[1] = tmp1+(number1*tmp3*(3*pow(2,(2./3.))*a[3]*getTerm(a)) - 
		   .5*tmp2*(number1*getTerm(a)));
  
  sol_t[2] = tmp1+(number2*tmp3*(3*pow(2,(2./3.))*a[3]*getTerm(a)) - 
		   .5*tmp2*(number2*getTerm(a)));
  
  sol[0] = sol_t[0].real();
  sol[1] = sol_t[1].real();
  sol[2] = sol_t[2].real();
  return ((sol_t[0].imag() != 0)||(sol_t[1].imag() != 0)
	  ||(sol_t[2].imag() != 0));
}

int VerletWeis::Solve3(const vector<double> & a, vector< complex <double> > & sol) 
{
  vector< complex<double> > sol_t(3);
  double tmp1 = -a[2]/(3*a[3]);
  double tmp2 = 1/(3*pow(2,(1./3.))*a[3]);
  double tmp3 = pow(2,(1./3.))*(-a[2]*a[2]+3*a[3]*a[1]);
  complex <double> number1(1, sqrt(3.0));
  complex <double> number2(1, -sqrt(3.0));
  sol_t[0] = tmp1-tmp3/(3*a[3]*getTerm(a))+tmp2*getTerm(a);
  
  sol_t[1] = tmp1+(number1*(-a[2]*a[2]+3*a[3]*a[1])/(3*pow(2,(2./3.))
						     *a[3]*getTerm(a)) - 
		   .5*tmp2*(number2*getTerm(a)));
  
  sol_t[2] = tmp1+(number2*(-a[2]*a[2]+3*a[3]*a[1])/(3*pow(2,(2./3.))*a[3]*getTerm(a)) - 
		   .5*tmp2*(number1*getTerm(a)));
  sol[0] = sol_t[0];
  sol[1] = sol_t[1];
  sol[2] = sol_t[2];
  return ((sol_t[0].imag() != 0)||(sol_t[1].imag() != 0)
	  ||(sol_t[2].imag() != 0));
}

complex<double> VerletWeis::getSquareroo(const vector<double> &a)
{
  //double tmp1, tmp2, tmp3;
  double tmp1 = 4*(pow((-pow(a[2],2)+3*a[3]*a[1]), 3)) ;
  double tmp2 = pow((-2*pow(a[2] ,3) +9*a[3]*a[2]*a[1] - 27*a[3]*a[3]*a[0]),2);
  return sqrt((complex<double>)(tmp1+tmp2));
  
}

complex<double> VerletWeis::getTerm(const vector<double> &a)
{
  complex<double> tmp1 = (-2*pow(a[2] ,3) +9*a[3]*a[2]*a[1] - 
			  27*a[3]*a[3]*a[0]) + getSquareroo(a);
 
  if(tmp1 == 0.0) 
    return tmp1;
  else
  return pow(tmp1,(complex<double>)(1./3.)) ;
}

double VerletWeis::Integrate(double *func, int nu, double mesh)
{
  double dummy=0, integral;
  int i;
 
  for(i=0;i<nu;i++) {
    
      
    if(i==0) 	
      dummy+=68323.l*func[i]+68323.l*func[i+1]-9531.l*func[i+2]+1879.l*func[i+3]
	
	-191.l*func[i+4];
    
    else if(i==1) 
      dummy+=-9531.l*func[i-1]+68323.l*func[i]+68323.l*func[i+1]
	-9531.l*func[i+2]+1879.l*func[i+3]-191.l*func[i+4];
    
    else if(i==2)  
      dummy+=+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i]+68323.l*func[i+1]
	-9531.l*func[i+2]+1879.l*func[i+3]-191.l*func[i+4];
    
    else if(i==nu-4)
       dummy+=+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i]+68323.l*func[i+1]
	-9531.l*func[i+2]+1879.l*func[i+3]; 
    
    else if(i==nu-3)
       dummy+=+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i]+68323.l*func[i+1]
	-9531.l*func[i+2];

    else if(i==nu-2)
       dummy+=+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i]+68323.l*func[i+1];
    
    else if(i==nu-1)
       dummy+=+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i];
    
    else
      dummy+=-191.l*func[i-3]+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i]
      +68323.l*func[i+1]-9531.l*func[i+2]+1879.l*func[i+3]-191.l*func[i+4];
  

    
  }
  integral=dummy*mesh/120960.l; 
  
  return integral;
  
}

double VerletWeis::Integrate(const vector<double>& func, 
		 const int& nu,
		 const double& mesh)
{
  double dummy=0, integral;
  int i;
 
  for(i=0;i<nu;i++) {
    
      
    if(i==0) 	
      dummy+=68323.l*func[i]+68323.l*func[i+1]-9531.l*func[i+2]+1879.l*func[i+3]
	
	-191.l*func[i+4];
    
    else if(i==1) 
      dummy+=-9531.l*func[i-1]+68323.l*func[i]+68323.l*func[i+1]
	-9531.l*func[i+2]+1879.l*func[i+3]-191.l*func[i+4];
    
    else if(i==2)  
      dummy+=+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i]+68323.l*func[i+1]
	-9531.l*func[i+2]+1879.l*func[i+3]-191.l*func[i+4];
    
    else if(i==nu-4)
       dummy+=+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i]+68323.l*func[i+1]
	-9531.l*func[i+2]+1879.l*func[i+3]; 
    
    else if(i==nu-3)
       dummy+=+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i]+68323.l*func[i+1]
	-9531.l*func[i+2];

    else if(i==nu-2)
       dummy+=+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i]+68323.l*func[i+1];
    
    else if(i==nu-1)
       dummy+=+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i];
    
    else
      dummy+=-191.l*func[i-3]+1879.l*func[i-2]-9531.l*func[i-1]+68323.l*func[i]
      +68323.l*func[i+1]-9531.l*func[i+2]+1879.l*func[i+3]-191.l*func[i+4];
  

    
  }
  integral=dummy*mesh/120960.l; 
  
  return integral;
}

